/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2001 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */


#ifndef _NV_LINUX_H_
#define _NV_LINUX_H_

#include "nv.h"

#include <linux/config.h>
#include <linux/version.h>


#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
#  error This driver does not support pre-2.4 kernels!
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
#  define KERNEL_2_4
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#  error This driver does not support 2.5 kernels!
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 7, 0)
#  define KERNEL_2_6
#else
#  error This driver does not support development kernels!
#endif

#if defined (CONFIG_SMP) && !defined (__SMP__)
#define __SMP__
#endif

#if defined (CONFIG_MODVERSIONS) && !defined (MODVERSIONS)
#  define MODVERSIONS
#endif

#if defined (MODVERSIONS) && !defined (KERNEL_2_6)
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>

#include <linux/init.h>             /* module_init, module_exit         */
#include <linux/types.h>            /* pic_t, size_t, __u32, etc        */
#include <linux/errno.h>            /* error codes                      */
#include <linux/list.h>             /* circular linked list             */
#include <linux/stddef.h>           /* NULL, offsetof                   */
#include <linux/wait.h>             /* wait queues                      */

#include <linux/slab.h>             /* kmalloc, kfree, etc              */
#include <linux/vmalloc.h>          /* vmalloc, vfree, etc              */

#include <linux/poll.h>             /* poll_wait                        */
#include <linux/delay.h>            /* mdelay, udelay                   */

#ifdef KERNEL_2_6
#include <linux/sched.h>            /* suser(), capable() replacement   */
#include <linux/moduleparam.h>      /* module_param()                   */
#include <linux/smp_lock.h>         /* kernel_locked                    */
#include <asm/tlbflush.h>           /* flush_tlb(), flush_tlb_all()     */
#include <asm/kmap_types.h>         /* page table entry lookup          */
#endif

#include <linux/pci.h>              /* pci_find_class, etc              */
#include <linux/interrupt.h>        /* tasklets, interrupt helpers      */
#include <linux/timer.h>

#include <asm/system.h>             /* cli, sli, save_flags             */
#include <asm/io.h>                 /* ioremap, virt_to_phys            */
#include <asm/uaccess.h>            /* access_ok                        */
#include <asm/page.h>               /* PAGE_OFFSET                      */
#include <asm/pgtable.h>            /* pte bit definitions              */

#if defined(NVCPU_X86_64) && defined(KERNEL_2_6)
#include <linux/syscalls.h>         /* sys_ioctl()                      */
#include <linux/ioctl32.h>          /* register_ioctl32_conversion()    */
#endif

#if defined(NVCPU_X86_64) && defined(KERNEL_2_4)
#include <asm/ioctl32.h>            /* sys_ioctl() (ioctl32)            */
#endif

#include <linux/spinlock.h>
#include <asm/semaphore.h>
#include <linux/highmem.h>

#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif

#ifdef CONFIG_PM
#include <linux/pm.h>
#endif

#ifdef CONFIG_MTRR
#include <asm/mtrr.h>
#endif

#ifdef CONFIG_KDB
#include <linux/kdb.h>
#include <asm/kdb.h>
#endif

#if defined(CONFIG_X86_REMOTE_DEBUG)
#include <linux/gdb.h>
#endif

#if defined(CONFIG_KGDB)
#include <linux/kgdb.h>
#endif

#if defined (CONFIG_AGP) || defined (CONFIG_AGP_MODULE)
#define AGPGART
#include <linux/agp_backend.h>
#include <linux/agpgart.h>
#endif

#if defined(NVCPU_X86) || defined(NVCPU_X86_64)
#define NV_BUILD_NV_PAT_SUPPORT 1
#endif

#if defined(NV_BUILD_NV_PAT_SUPPORT)
#include "pat.h"
#endif

#if defined(NVCPU_X86)
#ifndef write_cr4
#define write_cr4(x) __asm__ ("movl %0,%%cr4" :: "r" (x));
#endif

#ifndef read_cr4
#define read_cr4()                                  \
 ({                                                 \
      unsigned int __cr4;                           \
      __asm__ ("movl %%cr4,%0" : "=r" (__cr4));     \
      __cr4;                                        \
  })
#endif

#ifndef wbinvd
#define wbinvd() __asm__ __volatile__("wbinvd" ::: "memory");
#endif
#endif /* defined(NVCPU_X86) */

#ifndef get_cpu
#define get_cpu() smp_processor_id()
#define put_cpu()
#endif

#if !defined (list_for_each)
#define list_for_each(pos, head) \
        for (pos = (head)->next; pos != (head); pos = (pos)->next)
#endif

#if !defined(pmd_large)
#define pmd_large(_pmd) \
    ((pmd_val(_pmd) & (_PAGE_PSE|_PAGE_PRESENT)) == (_PAGE_PSE|_PAGE_PRESENT))
#endif

#ifdef EXPORT_NO_SYMBOLS
/* don't clutter the kernel namespace with our symbols */
EXPORT_NO_SYMBOLS;
#endif

#if !defined(DEBUG) && defined(__GFP_NOWARN)
#define NV_GFP_KERNEL (GFP_KERNEL | __GFP_NOWARN)
#define NV_GFP_ATOMIC (GFP_ATOMIC | __GFP_NOWARN)
#else
#define NV_GFP_KERNEL (GFP_KERNEL)
#define NV_GFP_ATOMIC (GFP_ATOMIC)
#endif

#define NV_GFP_DMA32 (NV_GFP_KERNEL)
#define CACHE_FLUSH()  asm volatile("wbinvd":::"memory")
#define WRITE_COMBINE_FLUSH() asm volatile("sfence":::"memory")


/* add support for iommu.
 * on x86_64 platforms, this uses the gart to remap pages that are > 32-bits
 * to < 32-bits.
 */
#if defined(NVCPU_X86_64)
#define NV_SG_MAP_BUFFERS 1
#endif

/* add support for software i/o tlb support.
 * normally, you'd expect this to be transparent, but unfortunately this is not
 * the case. for starters, the sw io tlb is a pool of pre-allocated pages that
 * are < 32-bits. when we ask to remap a page through this sw io tlb, we are
 * returned one of these pages, which means we have 2 different pages, rather
 * than 2 mappings to the same page. secondly, this pre-allocated pool is very
 * tiny, and the kernel panics when it is exhausted. try to warn the user that
 * they need to boost the size of their pool.
 */
#if defined(CONFIG_SWIOTLB)
#define NV_SWIOTLB 1
extern int nv_swiotlb;
#endif

/*
 * early 2.6 kernels changed their swiotlb codepath, running into a
 * latent bug that returns virtual addresses when it should return
 * physical addresses. we try to gracefully account for that, by 
 * comparing the returned address to what should be it's virtual
 * equivalent. this should hopefully account for when the bug is 
 * fixed in the core kernel.
 */
#if defined(NV_SWIOTLB) && defined(KERNEL_2_6)
#define NV_FIXUP_SWIOTLB_VIRT_ADDR_BUG(dma_addr) \
    if ((dma_addr) == ((dma_addr) | PAGE_OFFSET)) \
        (dma_addr) = __pa((dma_addr))
#else
#define NV_FIXUP_SWIOTLB_VIRT_ADDR_BUG(dma_addr)
#endif

#ifndef NVWATCH

/* various memory tracking/debugging techniques
 * disabled for retail builds, enabled for debug builds
 */

// allow an easy way to convert all debug printfs related to memory
// management back and forth between 'info' and 'errors'
#if defined(NV_DBG_MEM)
#define NV_DBG_MEMINFO NV_DBG_ERRORS
#else
#define NV_DBG_MEMINFO NV_DBG_INFO
#endif

#ifdef DEBUG
#define NV_ENABLE_MEM_TRACKING 1
#endif

#if NV_ENABLE_MEM_TRACKING
#define NV_MEM_TRACKING_PAD_SIZE(size)   ((size) += sizeof(void *))
#define NV_MEM_TRACKING_HIDE_SIZE(ptr, size)            \
    if ((ptr) && *(ptr)) {                              \
        U008 *__ptr;                                    \
        *(unsigned long *) *(ptr) = (size);             \
        __ptr = *(ptr); __ptr += sizeof(void *);        \
        *(ptr) = (void *) __ptr;                        \
    }
#define NV_MEM_TRACKING_RETRIEVE_SIZE(ptr, size)        \
    {                                                   \
        U008 *__ptr = (ptr); __ptr -= sizeof(void *);   \
        (ptr) = (void *) __ptr;                         \
        size = *(unsigned long *) (ptr);                \
    }
#else
#define NV_MEM_TRACKING_PAD_SIZE(size)
#define NV_MEM_TRACKING_HIDE_SIZE(ptr, size)
#define NV_MEM_TRACKING_RETRIEVE_SIZE(ptr, size)  ((size) = 0)
#endif


/* poor man's memory allocation tracker.
 * main intention is just to see how much memory is being used to recognize
 * when memory usage gets out of control or if memory leaks are happening
 */

/* keep track of memory usage */
#if NV_ENABLE_MEM_TRACKING
extern int vm_usage;
extern int km_usage;
extern int fp_usage;

/* print out a running tally of memory allocation amounts, disabled by default */
// #define POOR_MANS_MEM_CHECK 1


/* slightly more advanced memory allocation tracker.
 * track who's allocating memory and print out a list of currently allocated
 * memory at key points in the driver
 */

#define MEMDBG_ALLOC(a,b) (a = kmalloc(b, NV_GFP_ATOMIC))
#define MEMDBG_FREE(a)    (kfree(a))

#include "nv-memdbg.h"

#undef MEMDBG_ALLOC
#undef MEMDBG_FREE

extern struct mem_track_t *vm_list;
extern struct mem_track_t *km_list;
extern struct mem_track_t *fp_list;

/* print out list of memory allocations */
/* default to enabled for now */
#define LIST_MEM_CHECK 1

/* decide which memory types to apply mem trackers to */
#define VM_CHECKER 1
#define KM_CHECKER 1
#define FP_CHECKER 1

#endif  /* NV_ENABLE_MEM_TRACKING */

#define VM_PRINT(str, args...)
#define KM_PRINT(str, args...)
#define FP_PRINT(str, args...)

#define VM_ADD_MEM(a,b,c,d)
#define VM_FREE_MEM(a,b,c,d)
#define KM_ADD_MEM(a,b,c,d)
#define KM_FREE_MEM(a,b,c,d)
#define FP_ADD_MEM(a,b,c,d)
#define FP_FREE_MEM(a,b,c,d)

#define VM_ALLOC_RECORD(a,b,c)
#define VM_FREE_RECORD(a,b,c)
#define KM_ALLOC_RECORD(a,b,c)
#define KM_FREE_RECORD(a,b,c)


#if defined(VM_CHECKER)
#  if defined(POOR_MANS_MEM_CHECK)
#    undef  VM_PRINT
#    define VM_PRINT(str, args...)   printk(str, ##args)
#  endif
#  if defined(LIST_MEM_CHECK)
#    undef  VM_ADD_MEM
#    define VM_ADD_MEM(a,b,c,d)      nv_add_mem(&vm_list, a, b, c, d)
#    undef  VM_FREE_MEM
#    define VM_FREE_MEM(a,b,c,d)     nv_free_mem(&vm_list, a, b, c, d)
#  endif
#  undef VM_ALLOC_RECORD
#  define VM_ALLOC_RECORD(ptr, size, name) \
      if (ptr) { \
            vm_usage += size; \
            VM_PRINT("%s (0x%x: 0x%x): vm_usage is now 0x%x\n", \
                name, ptr, size, vm_usage); \
            VM_ADD_MEM(ptr, size, __FILE__, __LINE__); \
        }
#  undef VM_FREE_RECORD
#  define VM_FREE_RECORD(ptr, size, name) \
        if (ptr) { \
            vm_usage -= size; \
            VM_PRINT("%s (0x%x: 0x%x): vm_usage is now 0x%x\n", \
                name, ptr, size, vm_usage); \
            VM_FREE_MEM(ptr, size, __FILE__, __LINE__); \
        }
#endif

#if defined(KM_CHECKER)
#  if defined(POOR_MANS_MEM_CHECK)
#    undef  KM_PRINT
#    define KM_PRINT(str, args...)   printk(str, ##args)
#  endif
#  if defined(LIST_MEM_CHECK)
#    undef  KM_ADD_MEM
#    define KM_ADD_MEM(a,b,c,d)      nv_add_mem(&km_list, a, b, c, d)
#    undef  KM_FREE_MEM
#    define KM_FREE_MEM(a,b,c,d)     nv_free_mem(&km_list, a, b, c, d)
#  endif
#  undef KM_ALLOC_RECORD
#  define KM_ALLOC_RECORD(ptr, size, name) \
      if (ptr) { \
            km_usage += size; \
            KM_PRINT("%s (0x%x: 0x%x): km_usage is now 0x%x\n", \
                name, ptr, size, km_usage); \
            KM_ADD_MEM(ptr, size, __FILE__, __LINE__); \
        }
#  undef KM_FREE_RECORD
#  define KM_FREE_RECORD(ptr, size, name) \
        if (ptr) { \
            km_usage -= size; \
            KM_PRINT("%s (0x%x: 0x%x): km_usage is now 0x%x\n", \
                name, ptr, size, km_usage); \
            KM_FREE_MEM(ptr, size, __FILE__, __LINE__); \
        }
#endif

#if defined(FP_CHECKER)
#  if defined(POOR_MANS_MEM_CHECK)
#    undef  FP_PRINT
#    define FP_PRINT(str, args...)   printk(str, ##args)
#  endif
#if 0  // I had some problems tracking fp mem, so disable for now
#  if defined(LIST_MEM_CHECK)
#    undef  FP_ADD_MEM
#    define FP_ADD_MEM(a,b,c,d)      nv_add_mem(&fp_list, a, b, c, d)
#    undef  FP_FREE_MEM
#    define FP_FREE_MEM(a,b,c,d)     nv_free_mem(&fp_list, a, b, c, d)
#  endif
#endif
#endif

/* NV_VMALLOC has to allocate memory under the 4GB mark suitable for
 * DMA use with the hardware.
 */
#define NV_VMALLOC(ptr, size, cached)                                   \
    {                                                                   \
        pgprot_t __prot = (cached) ? PAGE_KERNEL : PAGE_KERNEL_NOCACHE; \
        (ptr) = __vmalloc(size, GFP_KERNEL, __prot);                    \
        VM_ALLOC_RECORD(ptr, size, "vm_vmalloc");                       \
    }

#define NV_VFREE(ptr, size)                         \
    {                                               \
        VM_FREE_RECORD(ptr, size, "vm_vmalloc");    \
        vfree((void *) (ptr));                      \
    }

#define NV_IOREMAP(ptr, physaddr, size) \
    { \
        (ptr) = ioremap(physaddr, size); \
        VM_ALLOC_RECORD(ptr, size, "vm_ioremap"); \
    }

#define NV_IOREMAP_NOCACHE(ptr, physaddr, size) \
    { \
        (ptr) = ioremap_nocache(physaddr, size); \
        VM_ALLOC_RECORD(ptr, size, "vm_ioremap_nocache"); \
    }

#define NV_IOUNMAP(ptr, size) \
    { \
        VM_FREE_RECORD(ptr, size, "vm_iounmap"); \
        iounmap(ptr); \
    }

/* only use this because GFP_KERNEL may sleep..
 * GFP_ATOMIC is ok, it won't sleep
 */
#define NV_KMALLOC(ptr, size) \
    { \
        (ptr) = kmalloc(size, NV_GFP_KERNEL); \
        KM_ALLOC_RECORD(ptr, size, "km_alloc"); \
    }

#define NV_KMALLOC_ATOMIC(ptr, size) \
    { \
        (ptr) = kmalloc(size, NV_GFP_ATOMIC); \
        KM_ALLOC_RECORD(ptr, size, "km_alloc_atomic"); \
    }  


#define NV_KFREE(ptr, size) \
    { \
        KM_FREE_RECORD(ptr, size, "km_free"); \
        kfree((void *) (ptr)); \
    }

#define NV_GET_FREE_PAGES(ptr, order) \
    { \
        (ptr) = __get_free_pages(NV_GFP_DMA32, order); \
    }
        
#define NV_FREE_PAGES(ptr, order) \
    { \
        free_pages(ptr, order); \
    }

#define NV_KMEM_CACHE_CREATE(kmem_cache, name, type)            \
    {                                                           \
        kmem_cache = kmem_cache_create(name, sizeof(type),      \
                        0, 0, NULL, NULL);                      \
    } 

#define NV_KMEM_CACHE_DESTROY(kmem_cache)                       \
    {                                                           \
        kmem_cache_destroy(kmem_cache);                         \
        kmem_cache = NULL;                                      \
    } 

#define NV_KMEM_CACHE_ALLOC(ptr, kmem_cache, type)              \
    {                                                           \
        (ptr) = kmem_cache_alloc(kmem_cache, GFP_KERNEL);       \
        KM_ALLOC_RECORD(ptr, sizeof(type), "km_cache_alloc");   \
    } 

#define NV_KMEM_CACHE_FREE(ptr, type, kmem_cache)               \
    {                                                           \
        KM_FREE_RECORD(ptr, sizeof(type), "km_cache_free");     \
        kmem_cache_free(kmem_cache, ptr);                       \
    } 

#if defined(NV_SG_MAP_BUFFERS) /* Linux/x86-64, only */
#if defined(NV_VMAP_2_PRESENT)

#define NV_VMAP(ptr, pages, count, cached)                              \
    {                                                                   \
        (ptr) = (unsigned long)vmap(pages, count);                      \
        VM_ALLOC_RECORD((void *)ptr, (count) * PAGE_SIZE, "vm_vmap");   \
    }

#elif defined(NV_VMAP_4_PRESENT)

#ifndef VM_MAP
#define VM_MAP  0
#endif

#define NV_VMAP(ptr, pages, count, cached)                              \
    {                                                                   \
        pgprot_t __prot = (cached) ? PAGE_KERNEL : PAGE_KERNEL_NOCACHE; \
        (ptr) = (unsigned long)vmap(pages, count, VM_MAP, __prot);      \
        VM_ALLOC_RECORD((void *)ptr, (count) * PAGE_SIZE, "vm_vmap");   \
    }

#else
#error "kernel function vmap(), which is necessary for x86_64 appears to be unavailable in this kernel!"
#endif

#define NV_VUNMAP(ptr, count)                                           \
    {                                                                   \
        VM_FREE_RECORD((void *)ptr, (count) * PAGE_SIZE, "vm_vmap");    \
        vunmap((void *)(ptr));                                          \
    }

#endif /* NV_SG_MAP_BUFFERS */

#endif /* !defined NVWATCH */

#if defined (KERNEL_2_4)
#  define NV_IS_SUSER()                 suser()
#  define NV_PCI_DEVICE_NAME(dev)       ((dev)->name)
#  define NV_NUM_CPUS()                 smp_num_cpus
#  define NV_CLI()                      __cli()
#  define NV_SAVE_FLAGS(eflags)         __save_flags(eflags)
#  define NV_RESTORE_FLAGS(eflags)      __restore_flags(eflags)
#  define NV_MAY_SLEEP()                (!in_interrupt())
#  define NV_MODULE_PARAMETER(x)        MODULE_PARM(x, "i")
#endif

#if defined (KERNEL_2_6)
#  define NV_IS_SUSER()                 capable(CAP_SYS_ADMIN)
#  define NV_PCI_DEVICE_NAME(dev)       ((dev)->pretty_name)
#  define NV_NUM_CPUS()                 num_online_cpus()
#  define NV_CLI()                      local_irq_disable()
#  define NV_SAVE_FLAGS(eflags)         local_save_flags(eflags)
#  define NV_RESTORE_FLAGS(eflags)      local_irq_restore(eflags)
#  define NV_MAY_SLEEP()                (!in_interrupt() && !in_atomic())
#  define NV_MODULE_PARAMETER(x)        module_param(x, int, 0)

   // the following macro causes problems when used in the same module
   // as module_param(); undef it so we don't accidentally mix the two
#  undef  MODULE_PARM
#endif

   // Earlier 2.4.x kernels don't have pci_disable_device()
#ifdef NV_PCI_DISABLE_DEVICE_PRESENT
#define NV_PCI_DISABLE_DEVICE(dev)      pci_disable_device(dev)
#else
#define NV_PCI_DISABLE_DEVICE(dev)
#endif

#ifdef NV_SIGNAL_STRUCT_RLIM
/* per-process rlimit settings */
#define NV_TASK_STRUCT_RLIM(current)  ((current)->signal->rlim)
#else
/* per-thread rlimit settings */
#define NV_TASK_STRUCT_RLIM(current)  ((current)->rlim)
#endif

/* common defines */
#define GET_MODULE_SYMBOL(mod,sym)    (const void *) inter_module_get(sym)
#define PUT_MODULE_SYMBOL(sym)        inter_module_put((char *) sym)

#define NV_GET_PAGE_STRUCT(phys_page) virt_to_page(__va(phys_page))
#define NV_VMA_PGOFF(vma)             ((vma)->vm_pgoff)
#define NV_VMA_SIZE(vma)              ((vma)->vm_end - (vma)->vm_start)
#define NV_VMA_OFFSET(vma)            ((vma)->vm_pgoff << PAGE_SHIFT)
#define NV_VMA_PRIVATE(vma)           ((vma)->vm_private_data)

#define NV_DEVICE_NUMBER(x)           minor((x)->i_rdev)
#define NV_IS_CONTROL_DEVICE(x)       (minor((x)->i_rdev) == 255)

#define NV_PCI_RESOURCE_START(dev, bar) pci_resource_start(dev, (bar))
#define NV_PCI_RESOURCE_SIZE(dev, bar)  pci_resource_len(dev, (bar))
#define NV_PCI_RESOURCE_FLAGS(dev, bar) pci_resource_flags(dev, (bar))
#define NV_PCI_RESOURCE_VALID(dev, bar) \
    (NV_PCI_RESOURCE_START(dev, bar) != 0 && NV_PCI_RESOURCE_SIZE(dev, bar) != 0)

#define NV_PCI_BUS_NUMBER(dev)        (dev)->bus->number
#define NV_PCI_DEVFN(dev)             (dev)->devfn
#define NV_PCI_SLOT_NUMBER(dev)       PCI_SLOT(NV_PCI_DEVFN(dev))

#ifdef NV_PCI_GET_CLASS_PRESENT
#define NV_PCI_DEV_PUT(dev)                    pci_dev_put(dev)
#define NV_PCI_GET_DEVICE(vendor,device,from)  pci_get_device(vendor,device,from)
#define NV_PCI_GET_SLOT(bus,devfn)                                       \
   ({                                                                    \
        struct pci_dev *__dev = NULL;                                    \
        while ((__dev = pci_get_device(PCI_ANY_ID, PCI_ANY_ID, __dev)))  \
        {                                                                \
            if (NV_PCI_BUS_NUMBER(__dev) == bus                          \
                    && NV_PCI_DEVFN(__dev) == devfn) break;              \
        }                                                                \
        __dev;                                                           \
    })
#define NV_PCI_GET_CLASS(class,from)           pci_get_class(class,from)
#else
#define NV_PCI_DEV_PUT(dev)
#define NV_PCI_GET_DEVICE(vendor,device,from)  pci_find_device(vendor,device,from)
#define NV_PCI_GET_SLOT(bus,devfn)             pci_find_slot(bus,devfn)
#define NV_PCI_GET_CLASS(class,from)           pci_find_class(class,from)
#endif

#if defined(DEBUG)
#define NV_PRINT_AT(at)                                                    \
    nv_printf(NV_DBG_ERRORS, "nvidia lost alloc_t: %4d  %2d  %05x 0x%08x 0x%08x\n", \
        at->num_pages, at->usage_count, \
        at->flags, at->page_table, at->key_mapping)
#else
#define NV_PRINT_AT(at)
#endif

// acpi support has been back-ported to the 2.4 kernel, but the 2.4 driver
// model is not sufficient for full acpi support. it may work in some cases,
// but not enough for us to officially support this configuration.
#if defined(CONFIG_ACPI) && defined(KERNEL_2_6)
#define NV_PM_SUPPORT_ACPI
#endif

#if defined(CONFIG_APM) || defined(CONFIG_APM_MODULE)
#define NV_PM_SUPPORT_APM
#endif

#ifndef minor
# define minor(x) MINOR(x)
#endif

#ifndef IRQ_RETVAL
typedef void irqreturn_t;
#define IRQ_RETVAL(a)
#endif

#ifndef PCI_CAP_ID_EXP
#define PCI_CAP_ID_EXP 0x10
#endif

#if defined(NV_REMAP_PFN_RANGE_PRESENT)
#define NV_REMAP_PAGE_RANGE(from, offset, x...) \
     remap_pfn_range(vma, from, ((offset) >> PAGE_SHIFT), x)
#elif defined(NV_REMAP_PAGE_RANGE_5_PRESENT)
#define NV_REMAP_PAGE_RANGE(x...) remap_page_range(vma, x)
#elif defined(NV_REMAP_PAGE_RANGE_4_PRESENT)
#define NV_REMAP_PAGE_RANGE(x...) remap_page_range(x)
#else
#warning "conftest.sh failed, assuming remap_page_range(4)!"
#define NV_REMAP_PAGE_RANGE(x...) remap_page_range(x)
#endif


#define NV_PGD_OFFSET(address, kernel, mm)              \
   ({                                                   \
        pgd_t *__pgd;                                   \
        if (!kernel)                                    \
            __pgd = pgd_offset(mm, address);            \
        else                                            \
            __pgd = pgd_offset_k(address);              \
        __pgd;                                          \
    })

#define NV_PGD_PRESENT(pgd)                             \
   ({                                                   \
         if ((pgd != NULL) &&                           \
             (pgd_bad(*pgd) || pgd_none(*pgd)))         \
            /* static */ pgd = NULL;                    \
         pgd != NULL;                                   \
    })

#if defined(pmd_offset_map)
#define NV_PMD_OFFSET(address, pgd)                     \
   ({                                                   \
        pmd_t *__pmd;                                   \
        __pmd = pmd_offset_map(pgd, address);           \
   })
#define NV_PMD_UNMAP(pmd) pmd_unmap(pmd);
#else
#if defined(PUD_SHIFT) /* 4-level pgtable */
#define NV_PMD_OFFSET(address, pgd)                     \
   ({                                                   \
        pmd_t *__pmd = NULL;                            \
        pud_t *__pud;                                   \
        __pud = pud_offset(pgd, address);               \
        if ((__pud != NULL) &&                          \
            !(pud_bad(*__pud) || pud_none(*__pud)))     \
            __pmd = pmd_offset(__pud, address);         \
        __pmd;                                          \
    })
#else /* 3-level pgtable */
#define NV_PMD_OFFSET(address, pgd)                     \
   ({                                                   \
        pmd_t *__pmd;                                   \
        __pmd = pmd_offset(pgd, address);               \
    })
#endif
#define NV_PMD_UNMAP(pmd)
#endif

#define NV_PMD_PRESENT(pmd)                             \
   ({                                                   \
        if ((pmd != NULL) &&                            \
            (pmd_bad(*pmd) || pmd_none(*pmd)))          \
        {                                               \
            NV_PMD_UNMAP(pmd);                          \
            pmd = NULL; /* mark invalid */              \
        }                                               \
        pmd != NULL;                                    \
    })

#if defined(pte_offset_atomic)
#define NV_PTE_OFFSET(address, pmd)                     \
   ({                                                   \
        pte_t *__pte;                                   \
        __pte = pte_offset_atomic(pmd, address);        \
        NV_PMD_UNMAP(pmd); __pte;                       \
    })
#define NV_PTE_UNMAP(pte) pte_kunmap(pte);
#elif defined(pte_offset)
#define NV_PTE_OFFSET(address, pmd)                     \
   ({                                                   \
        pte_t *__pte;                                   \
        __pte = pte_offset(pmd, address);               \
        NV_PMD_UNMAP(pmd); __pte;                       \
    })
#define NV_PTE_UNMAP(pte)
#else
#define NV_PTE_OFFSET(address, pmd)                     \
   ({                                                   \
        pte_t *__pte;                                   \
        __pte = pte_offset_map(pmd, address);           \
        NV_PMD_UNMAP(pmd); __pte;                       \
    })
#define NV_PTE_UNMAP(pte) pte_unmap(pte);
#endif

#define NV_PTE_PRESENT(pte)                             \
   ({                                                   \
        if ((pte != NULL) && !pte_present(*pte))        \
        {                                               \
            NV_PTE_UNMAP(pte);                          \
            pte = NULL; /* mark invalid */              \
        }                                               \
        pte != NULL;                                    \
    })

#define NV_PTE_VALUE(pte)                               \
   ({                                                   \
        unsigned long __pte_value = pte_val(*pte);      \
        NV_PTE_UNMAP(pte);                              \
        __pte_value;                                    \
    })


#define NV_PAGE_ALIGN(addr)             ( ((addr) + PAGE_SIZE - 1) / PAGE_SIZE)
#define NV_MASK_OFFSET(addr)            ( (addr) & (PAGE_SIZE - 1) )

#if defined(NVCPU_X86) || defined(NVCPU_X86_64)
/* this isn't defined in some older kernel header files */
#define NV_CPU_INTERRUPT_FLAGS_BIT (1<<9)
#else
#error define NV_CPU_INTERRUPT_FLAGS_BIT
#endif

static inline int NV_IRQL_IS_RAISED(void)
    {
        unsigned long int eflags;
        NV_SAVE_FLAGS(eflags);
        return !(eflags & NV_CPU_INTERRUPT_FLAGS_BIT);
    }
 
static inline int nv_calc_order(unsigned int size)
    {
        int order = 0;
        while ( ((1 << order) * PAGE_SIZE) < (size))
        {
            order++;
        }
        return order;
    }

/* mark memory UC-, rather than UC (don't use _PAGE_PWT) */
static inline pgprot_t pgprot_noncached_weak(pgprot_t old_prot)
    {
        pgprot_t new_prot = old_prot;
        if (boot_cpu_data.x86 > 3)
            new_prot = __pgprot(pgprot_val(old_prot) | _PAGE_PCD);
        return new_prot;
    }

#if !defined (pgprot_noncached)
static inline pgprot_t pgprot_noncached(pgprot_t old_prot)
    {
        pgprot_t new_prot = old_prot;
        if (boot_cpu_data.x86 > 3)
            new_prot = __pgprot(pgprot_val(old_prot) | _PAGE_PCD | _PAGE_PWT);
        return new_prot;
    }
#endif

#if defined(NV_BUILD_NV_PAT_SUPPORT) && !defined (pgprot_writecombined)
static inline pgprot_t pgprot_writecombined(pgprot_t old_prot)
    {
        pgprot_t new_prot = old_prot;
        if (boot_cpu_data.x86 > 3)
        {
            pgprot_val(old_prot) &= ~(_PAGE_PCD | _PAGE_PWT);
            new_prot = __pgprot(pgprot_val(old_prot) | _PAGE_WRTCOMB);
        }
        return new_prot;
    }
#endif

#if !defined(page_to_pfn)
#define page_to_pfn(page)  ((page) - mem_map)
#endif

#if !defined(pfn_to_page)
#define pfn_to_page(pfn) (mem_map + (pfn))
#endif

/*
 * An allocated bit of memory using NV_MEMORY_ALLOCATION_OFFSET
 *   looks like this in the driver
 */

typedef struct nv_pte_t {
    unsigned long   phys_addr;
    unsigned long   virt_addr;
    dma_addr_t      dma_addr;
#ifdef NV_SG_MAP_BUFFERS
    struct scatterlist sg_list;
#endif
#ifdef CONFIG_SWIOTLB
    unsigned long   orig_phys_addr;
    unsigned long   orig_virt_addr;
#endif
} nv_pte_t;

typedef struct nv_alloc_s {
    struct nv_alloc_s *next;    
    atomic_t       usage_count;
    unsigned int   flags;
    unsigned int   num_pages;
    unsigned int   order;
    unsigned int   size;
    nv_pte_t     **page_table;          /* list of physical pages allocated */
    void          *key_mapping;         /* mapping used as a key for finding this nv_alloc_t */
                                        /*   may be the same as page_table                   */
    unsigned int   class;
    void          *priv_data;
} nv_alloc_t;


#define NV_ALLOC_TYPE_PCI      (1<<0)
#define NV_ALLOC_TYPE_AGP      (1<<1)
#define NV_ALLOC_TYPE_CONTIG   (1<<2)
#define NV_ALLOC_TYPE_KERNEL   (1<<3)
#define NV_ALLOC_TYPE_VMALLOC  (1<<4)
#define NV_ALLOC_TYPE_VMAP     (1<<5)

#define NV_ALLOC_MAPPING_SHIFT      16
#define NV_ALLOC_MAPPING(flags)     (((flags)>>NV_ALLOC_MAPPING_SHIFT)&0xff)
#define NV_ALLOC_ENC_MAPPING(flags) ((flags)<<NV_ALLOC_MAPPING_SHIFT)

#define NV_ALLOC_MAPPING_CACHED(flags) (NV_ALLOC_MAPPING(flags) == NV_MEMORY_CACHED)

#define NV_ALLOC_MAPPING_AGP(flags)     ((flags) & NV_ALLOC_TYPE_AGP)
#define NV_ALLOC_MAPPING_CONTIG(flags)  ((flags) & NV_ALLOC_TYPE_CONTIG)
#define NV_ALLOC_MAPPING_VMALLOC(flags) ((flags) & NV_ALLOC_TYPE_VMALLOC)
#define NV_ALLOC_MAPPING_VMAP(flags)    ((flags) & NV_ALLOC_TYPE_VMAP)

static inline U032 nv_alloc_init_flags(int cached, int agp, int contig, int kernel)
{
    U032 flags = NV_ALLOC_ENC_MAPPING(cached);
    if (agp)    flags |= NV_ALLOC_TYPE_AGP;
    else        flags |= NV_ALLOC_TYPE_PCI;
    if (kernel) flags |= NV_ALLOC_TYPE_KERNEL; 
#if defined(NV_SG_MAP_BUFFERS)
    if (kernel && !contig) flags |= NV_ALLOC_TYPE_VMAP;
#else
    if (kernel && !contig) flags |= NV_ALLOC_TYPE_VMALLOC;
#endif
    if (contig && !agp) flags |= NV_ALLOC_TYPE_CONTIG;
    return flags;
}

/* linux-specific version of old nv_state_t */
/* this is a general os-specific state structure. the first element *must* be
   the general state structure, for the generic unix-based code */
typedef struct {
    nv_state_t nv_state;
    atomic_t usage_count;

    struct pci_dev *dev;
    void *agp_bridge;
    nv_alloc_t *alloc_queue;

    /* keep track of any pending bottom halfes */
    struct tasklet_struct tasklet;

    /* get a timer callback every second */
    struct timer_list rc_timer;

    /* per-device locking mechanism for access to core rm */
    spinlock_t rm_lock;
    int rm_lock_cpu;
    int rm_lock_count;

    /* lock for linux-specific data, not used by core rm */
    struct semaphore ldata_lock;

    /* lock for linux-specific alloc queue */
    struct semaphore at_lock;
} nv_linux_state_t;

extern int nv_pat_enabled;

/*
 * file-private data
 * hide a pointer to our data structures in a file-private ptr
 * there are times we need to grab this data back from the file
 * data structure..
 */

#define NV_EVENT_FIFO_SIZE 6

typedef struct
{
    void *nvptr;
    U032 num_events;
    U032 put, get;
    spinlock_t fp_lock;
    wait_queue_head_t waitqueue;
    nv_event_t *event_fifo;     // fifo for storing events
} nv_file_private_t;

#define FILE_PRIVATE(filep)     ((filep)->private_data)

#define NV_GET_NVFP(filep)      ((nv_file_private_t *) FILE_PRIVATE(filep))

/* for the card devices */
#define NVL_FROM_FILEP(filep)   (NV_GET_NVFP(filep)->nvptr)

#define NV_GET_NVL_FROM_NV_STATE(nv) \
    ((nv_linux_state_t *) nv->os_state)

#define NV_STATE_PTR(nvl)   (&((nvl)->nv_state))


#define NV_ATOMIC_SET(data,val)         atomic_set(&(data), (val))
#define NV_ATOMIC_INC(data)             atomic_inc(&(data))
#define NV_ATOMIC_DEC(data)             atomic_dec(&(data))
#define NV_ATOMIC_DEC_AND_TEST(data)    atomic_dec_and_test(&(data))
#define NV_ATOMIC_READ(data)            atomic_read(&(data))

/*
 * AMD Athlon processors expose a subtle bug in the Linux
 * kernel, that may lead to AGP memory corruption. Recent
 * kernel versions had a workaround for this problem, but
 * 2.4.20 is the first kernel to address it properly. The
 * page_attr API provides the means to solve the problem. 
 */

/*
 * a BUG() is triggered on early 2.6 x86_64 kernels. the underlying
 * problem actually exists on many architectures and kernels, but
 * these are the only kernels that check the condition and trigger
 * a BUG(). note that this is a problem of the core kernel, not an
 * nvidia bug (and can still be triggered by agpgart). let's avoid
 * change_page_attr on those kernels.
 */
#if defined(NVCPU_X86_64) && defined(KERNEL_2_6) && \
        (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11))
#define NV_CHANGE_PAGE_ATTR_BUG_PRESENT 1
#endif

#if defined(NV_CHANGE_PAGE_ATTR_PRESENT) && !defined(NV_CHANGE_PAGE_ATTR_BUG_PRESENT)
static inline void NV_SET_PAGE_ATTRIB_UNCACHED(nv_pte_t *page_ptr)
    {
        struct page *page = virt_to_page(__va(page_ptr->phys_addr));
        change_page_attr(page, 1, PAGE_KERNEL_NOCACHE);
    }
static inline void NV_SET_PAGE_ATTRIB_CACHED(nv_pte_t *page_ptr)
    {
        struct page *page = virt_to_page(__va(page_ptr->phys_addr));
        change_page_attr(page, 1, PAGE_KERNEL);
    }
#else
#define NV_SET_PAGE_ATTRIB_UNCACHED(page_list)
#define NV_SET_PAGE_ATTRIB_CACHED(page_list)
#endif

static inline void NV_UNLOCK_PAGE(nv_pte_t *page_ptr)
    {
        unsigned long phys_addr;
#ifdef CONFIG_SWIOTLB
        if (page_ptr->orig_phys_addr) phys_addr = page_ptr->orig_phys_addr;
        else
#endif
            phys_addr = page_ptr->phys_addr;
        ClearPageReserved(NV_GET_PAGE_STRUCT(phys_addr));
    }

#if defined(NV_SG_MAP_BUFFERS)

static inline int nv_map_sg(struct pci_dev *dev, struct scatterlist *sg)
    {
        int ret;
#if defined(KERNEL_2_6) && defined(CONFIG_SWIOTLB)
        if (swiotlb)
            ret = swiotlb_map_sg(&dev->dev, sg, 1, PCI_DMA_BIDIRECTIONAL);
        else
#endif
            ret = pci_map_sg(dev, sg, 1, PCI_DMA_BIDIRECTIONAL);
        return ret;
    }

static inline void nv_unmap_sg(struct pci_dev *dev, struct scatterlist *sg)
    {
#if defined(KERNEL_2_6) && defined(CONFIG_SWIOTLB)
        if (swiotlb)
            swiotlb_unmap_sg(&dev->dev, sg, 1, PCI_DMA_BIDIRECTIONAL);
        else
#endif
            pci_unmap_sg(dev, sg, 1, PCI_DMA_BIDIRECTIONAL);
    }

#define NV_MAP_SG_MAX_RETRIES 16

static inline int nv_sg_map_buffer(
    struct pci_dev     *dev,
    nv_pte_t          **page_list,
    void               *base,
    unsigned int        num_pages
)
{
    struct scatterlist *sg_ptr = &page_list[0]->sg_list;
    unsigned int i;

    sg_ptr->page = virt_to_page(base);
    sg_ptr->offset = (unsigned long)base & ~PAGE_MASK;
    sg_ptr->length  = num_pages * PAGE_SIZE;

    i = NV_MAP_SG_MAX_RETRIES;
    do {
        if (nv_map_sg(dev, sg_ptr) == 0)
            return 1;

        if (sg_ptr->dma_address & ~PAGE_MASK)
        {
            struct scatterlist sg_tmp;
            nv_unmap_sg(dev, sg_ptr);

            memset(&sg_tmp, 0, sizeof(struct scatterlist));
            sg_tmp.page = sg_ptr->page;
            sg_tmp.offset = sg_ptr->offset;
            sg_tmp.length = 2048;

            if (nv_map_sg(dev, &sg_tmp) == 0)
                return 1;

            if (nv_map_sg(dev, sg_ptr) == 0)
            {
                nv_unmap_sg(dev, &sg_tmp);
                return 1;
            }

            nv_unmap_sg(dev, &sg_tmp);
        }
    } while (i-- && sg_ptr->dma_address & ~PAGE_MASK);

    if (sg_ptr->dma_address & ~PAGE_MASK)
    {
        nv_printf(NV_DBG_ERRORS,
            "NVRM: VM: nv_sg_map_buffer: failed to obtain aligned mapping\n");
        nv_unmap_sg(dev, sg_ptr);
        return 1;
    }


    // this is a bit of a hack to make contiguous allocations easier to handle
    // nv_sg_load below relies on the page_ptr addresses being filed in, as 
    // well as the sg_ptr having a valid dma_address. most allocations call
    // nv_sg_map_buffers page-by-page, but contiguous allocations will make
    // one call for the whole allocation. make sure we correctly propogate 
    // our dma_address through the rest of the sg_ptrs for these allocations.
    // note we start with index 1, since index 0 is already correct
    for (i = 1; i < num_pages; i++)
    {
        page_list[i]->sg_list.dma_address = sg_ptr->dma_address + (i * PAGE_SIZE);
    }

    return 0;
}

static inline int nv_sg_load(
    struct scatterlist *sg_ptr,
    nv_pte_t           *page_ptr
)
{
    page_ptr->dma_addr = sg_ptr->dma_address;

#if defined(NV_SWIOTLB)
    // with the sw io tlb, we've actually switched to different physical pages
    // wire in the new page's addresses, but save the original off to free later
    if (nv_swiotlb)
    {
        // note that we modify our local version, not the sg_ptr version that
        // will be returned to the swiotlb pool
        NV_FIXUP_SWIOTLB_VIRT_ADDR_BUG(page_ptr->dma_addr);
        page_ptr->orig_phys_addr = page_ptr->phys_addr;
        page_ptr->phys_addr      = page_ptr->dma_addr;
        page_ptr->orig_virt_addr = page_ptr->virt_addr;
        page_ptr->virt_addr      = (unsigned long) __va(page_ptr->dma_addr);
    }
#endif

    return 0;
}

// make sure we only unmap the page if it was really mapped through the iommu,
// in which case the dma_addr and phys_addr will not match.
static inline void nv_sg_unmap_buffer(
    struct pci_dev     *dev,
    struct scatterlist *sg_ptr,
    nv_pte_t           *page_ptr
)
{
#ifdef CONFIG_SWIOTLB
    // for sw io tlbs, dma_addr == phys_addr currently, so the check below fails
    // restore the original settings first, then the following check will work
    if (nv_swiotlb && page_ptr->orig_phys_addr)
    {
        page_ptr->phys_addr      = page_ptr->orig_phys_addr;
        page_ptr->virt_addr      = page_ptr->orig_virt_addr;
        page_ptr->orig_phys_addr = 0;
        page_ptr->orig_virt_addr = 0;
    }
#endif

    if (page_ptr->dma_addr != page_ptr->phys_addr)
    {
        nv_unmap_sg(dev, sg_ptr);
        page_ptr->dma_addr = 0;
    }
}
#endif  /* NV_SG_MAP_BUFFERS */

/*
 * Basic support for kgdb assertions.
 */
#if defined(CONFIG_X86_REMOTE_DEBUG)
#define NV_ASSERT(message, condition) KGDB_ASSERT(message, condition)
#else
#if defined(DEBUG)
#define NV_ASSERT(message, condition) \
if (!(condition)) { \
    nv_printf(NV_DBG_ERRORS, "NVRM: ASSERT: %s\n", message); \
    os_dbg_breakpoint(); \
} else
#else
#define NV_ASSERT(message, condition)
#endif /* DEBUG */
#endif

int nv_verify_page_mappings(unsigned long, unsigned int);

#endif  /* _NV_LINUX_H_ */
